/******************************************************************************
 * Copyright (c) 2005 Eric Petit
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <apr.h>
#include <apr_pools.h>
#include <libbtutil.h>

#define LIST_SIZE 20

static const char *bt_benc_errors[] = {
    "Success",
    "Bad Integer",
    "Bad List",
    "Bad Dictionary",
    "Uneven Dictionary",
    "Bad String",
    "Bad String Length",
    NULL
};

const char* bt_benc_error (int val) {
    return bt_benc_errors[val];
}

int bt_parse_bencode(apr_pool_t* pool, char* buf, bt_bencode* val, char** end) {
    char * p, * foo;

    /* So we only have to check once */
    if(!end)
        end = &foo;

    val->begin = buf;

    if( buf[0] == 'i' ) {
        /* Integer: i1242e */
        val->type  = BT_BENC_TYPE_INT;
        val->val.i = strtoll( &buf[1], &p, 10 );

        if( p == &buf[1] || p[0] != 'e' )
            return BT_BENC_BAD_INT;

        val->end = p + 1;
    } else if( buf[0] == 'l' || buf[0] == 'd' ) {
        /* List: l<item1><item2>e
           Dict: d<string1><item1><string2><item2>e
           A dictionary is just a special kind of list with an even
           count of items, and where even items are strings. */
        char * cur;
        char   is_dict;
        char   str_expected;

        is_dict          = ( buf[0] == 'd' );
        val->type        = is_dict ? BT_BENC_TYPE_DICT : BT_BENC_TYPE_LIST;
        val->val.l.alloc = LIST_SIZE;
        val->val.l.count = 0;
        val->val.l.vals  = apr_palloc(pool, LIST_SIZE * sizeof(bt_bencode));
        cur              = &buf[1];
        str_expected     = 1;
        
        while(cur[0] != 'e') {
            if(val->val.l.count == val->val.l.alloc) {
                /* grow our pool. APR doesn't have realloc() so we have
                 * to fake it. this means the old memory is left laying
                 * around, but pools for bdecodes should be temporary
                 * anyway.
                 */
                void* oldmem = val->val.l.vals;
                val->val.l.vals = apr_palloc(
                    pool, (val->val.l.alloc + LIST_SIZE) * sizeof(bt_bencode)
                );
                memcpy(
                    val->val.l.vals, oldmem,
                    val->val.l.alloc * sizeof(bt_bencode)
                );
                val->val.l.alloc += LIST_SIZE;
            }
            
            if(bt_parse_bencode(pool, cur, &val->val.l.vals[val->val.l.count], &p ))
                return BT_BENC_BAD_LIST;

            if(
                is_dict && str_expected &&
                val->val.l.vals[val->val.l.count].type != BT_BENC_TYPE_STR
            )
                return BT_BENC_BAD_DICT;

            str_expected = !str_expected;

            val->val.l.count++;
            cur = p;
        }

        if( is_dict && ( val->val.l.count & 1 ) )
            return BT_BENC_DICT_UNEVEN;

        val->end = cur + 1;
    } else {
        /* String: 12:whateverword */
        val->type    = BT_BENC_TYPE_STR;
        val->val.s.i = strtol( buf, &p, 10 );

        if(p == buf) return BT_BENC_BAD_STRING_LENGTH;
        if(p[0] != ':') return BT_BENC_BAD_STRING;
        
        val->val.s.s               = apr_palloc(pool, val->val.s.i + 1);
        val->val.s.s[val->val.s.i] = 0;
        memcpy( val->val.s.s, p + 1, val->val.s.i );

        val->end = p + 1 + val->val.s.i;
    }

    *end = val->end;

    return BT_BENC_SUCCESS;
}

bt_bencode* bt_find_bencode(bt_bencode * val, char * key) {
    int i;
    
    if(val->type != BT_BENC_TYPE_DICT)
        return NULL;
    
    for(i=0; i<val->val.l.count; i+=2 )
        if(!strcmp(val->val.l.vals[i].val.s.s,key))
            return &val->val.l.vals[i+1];

    return NULL;
}
