/******************************************************************************
 * Copyright (c) 2005 Eric Petit
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *****************************************************************************/

#include <libbtutil.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <malloc.h>
#include <stdlib.h>

#include <apr.h>
#include <apr_pools.h>
#include <apr_strings.h>
#include <apr_file_info.h>
#include <apr_file_io.h>

#define SF APR_SIZE_T_FMT

static char * strcatUTF8( apr_pool_t *, char *, char * );

static char errorstr[BT_SHORT_STRING];

static char *load_metainfo(apr_pool_t* p, const char* path) {
    char*               buf;
    apr_file_t*         file;
    apr_size_t          len;
    struct stat         sinfo;
    int                 rv;

    bzero(&sinfo, sizeof(sinfo));

    if(stat(path, &sinfo)) {
        fprintf(stderr, "Failed to stat %s\n", path);
        return NULL;
    }
    
    if(!S_ISREG(sinfo.st_mode)) {
        fprintf(stderr, "%s: Not a regular file.\n", path);
        return NULL;
    }
    
    if(sinfo.st_size > BT_MAX_METAINFO_SIZE) {
        fprintf(
            stderr, "%s: file to large (%llu bytes > %u bytes)",
            path, sinfo.st_size, BT_MAX_METAINFO_SIZE
        );
        return NULL;
    }
    
    /* Load the torrent file into our buffer */
    if((rv = apr_file_open(&file, path, APR_READ | APR_BINARY, APR_OS_DEFAULT, p)) != APR_SUCCESS) {
        fprintf(
            stderr, "Could not open file (%s): %s\n", path,
            apr_strerror(rv, errorstr, BT_SHORT_STRING)
        );
        return NULL;
    }
    
    len = sinfo.st_size;
    buf = apr_palloc(p, len);
    
    if((rv = apr_file_read(file, buf, &len)) != APR_SUCCESS) {
        fprintf(
            stderr, "Failed to read %llu bytes from %s: %s\n",
            sinfo.st_size, path,
            apr_strerror(rv, errorstr, BT_SHORT_STRING)
        );
        return NULL;
    }
    
    if(len != sinfo.st_size) {
        fprintf(
            stderr,
            "%s: expected %llu bytes, but got %"SF"!\n", path, sinfo.st_size, len
        );
        return NULL;
    }
    
    apr_file_close(file);
    
    return buf;
}

bt_metainfo* bt_parse_metainfo(apr_pool_t* p, const char* path) {
    char*               buf;
    bt_bencode          meta, *beInfo, *list, *val;
    char                *s, *s2, *s3;
    int                 i;
    apr_pool_t*         temp = NULL;
    bt_metainfo*        metainfo;
    int rv;
       
    if(apr_pool_create(&temp, p) != APR_SUCCESS) {
        fprintf(stderr, "Error creating metainfo's temporary pool!\n");
        goto err;
    }

    metainfo = apr_pcalloc(p, sizeof(bt_metainfo));
    strncpy(metainfo->torrent, path, sizeof(metainfo->torrent) - 1);

    if(!(buf = load_metainfo(temp, path)))
        goto err;
    
    /* Parse bencoded infos */
    if((rv = bt_parse_bencode(temp, buf, &meta, NULL)) != BT_BENC_SUCCESS) {
        fprintf(stderr, "Error while parsing bencoded data: %s\n", bt_benc_error(rv));
        goto err;
    }

    /* Get info hash */
    if(!(beInfo = bt_find_bencode(&meta, "info"))) {
        fprintf(stderr, "Could not find \"info\" dictionary\n");
        goto err;
    }
    
    BT_SHA1(
        (uint8_t *) beInfo->begin,
        (long) beInfo->end - (long) beInfo->begin, metainfo->hash
    );


    if(!(val = bt_find_bencode(&meta, "announce" )))
    {
        fprintf(stderr, "No \"announce\" entry\n");
        goto err;
    }
    
    /* Skip spaces */
    s3 = val->val.s.s;
    while( *s3 && *s3 == ' ' )
        s3++;

    /* Parse announce URL */
    
    if(strlen(s3) > sizeof(metainfo->tracker_address) - 1) {
        fprintf(stderr, "Too long announce url (%s)\n", s3);
        goto err;
    }
    
    if(strncmp( s3, "http://", 7)) {
        fprintf(stderr, "Invalid announce URL (%s)\n", s3);
       goto err;
    }
    
    s  = strchr( s3 + 7, ':' );
    s2 = strchr( s3 + 7, '/' );
    
    if(s && s < s2) {
        memcpy(metainfo->tracker_address, s3 + 7, (long)s - (long)s3 - 7);
        metainfo->tracker_port = atoi(s + 1);
    } else if(s2) {
        memcpy(metainfo->tracker_address , s3 + 7, (long)s2 - (long)s3 - 7);
        metainfo->tracker_port = 80;
    } else {
        fprintf(stderr, "Invalid announce URL (%s)\n", s3);
        goto err;
    }
    
    strncpy(
        metainfo->tracker_announce, s2, sizeof(metainfo->tracker_announce) - 1
    );

    /* Piece length */
    if(!(val = bt_find_bencode(beInfo, "piece length"))) {
        fprintf(stderr, "No \"piece length\" entry\n");
        goto err;
    }
    
    metainfo->piece_size = val->val.i;

    /* Hashes */
    val = bt_find_bencode(beInfo, "pieces");
    if(val->val.s.i % BT_INFOHASH_LEN) {
        fprintf(
            stderr, "Invalid \"piece\" string (size is %d)\n", val->val.s.i
        );
        goto err;
    }
    
    metainfo->piece_count = val->val.s.i / BT_INFOHASH_LEN;
    metainfo->pieces = apr_pmemdup(p, val->val.s.s, val->val.s.i);

    /* TODO add more tests so we don't crash on weird files */

    metainfo->total_size = 0;
    
    if((list = bt_find_bencode(beInfo, "files"))) {
        /* Multi-file mode */
        int j;
        char *name;

        val = bt_find_bencode(beInfo, "name");
        name = strcatUTF8(temp, "", val->val.s.s);
        strncpy(metainfo->name, name, sizeof(metainfo->name) - 1);

        metainfo->file_count = list->val.l.count;
        metainfo->files      = apr_pcalloc(
            p, metainfo->file_count * sizeof(bt_metainfo_finfo)
        );

        for(i=0; i<list->val.l.count; i++) {
            val = bt_find_bencode(&list->val.l.vals[i], "path");
            name = strcatUTF8(temp, "", metainfo->name);
            
            for(j=0; j<val->val.l.count; j++) {
                name = strcatUTF8(temp, name, "/");
                name = strcatUTF8(temp, name, val->val.l.vals[j].val.s.s);
                
                strncpy(
                    metainfo->files[i].name, name,
                    sizeof(metainfo->files[i].name) - 1
                );
            }
            val = bt_find_bencode(&list->val.l.vals[i], "length");
            metainfo->files[i].length  = val->val.i;
            metainfo->total_size       += val->val.i;
        }
    } else {
        /* Single-file mode */
        char* name;
        
        metainfo->file_count    = 1;
        metainfo->files         = apr_pcalloc(p, sizeof(bt_metainfo_finfo));

        val = bt_find_bencode(beInfo, "name");
        name = strcatUTF8(p, "", val->val.s.s);
        
        strncpy(
            metainfo->files[0].name, name,
            sizeof(metainfo->files[i].name) - 1
        );
        
        strncpy(metainfo->name, name, sizeof(metainfo->name) - 1);
        
        val = bt_find_bencode(beInfo, "length");
        metainfo->files[0].length  = val->val.i;
        metainfo->total_size       += val->val.i;
    }

    if(
        (uint64_t)metainfo->piece_count !=
        (metainfo->total_size + metainfo->piece_size - 1) / metainfo->piece_size
    ) {
        fprintf( stderr, "Size of hashes and files don't match\n" );
        goto err;
    }
    
    apr_pool_destroy(temp);
    return metainfo;
    
    err:
    if(temp) apr_pool_destroy(temp);
    if(p) apr_pool_destroy(p);
    return NULL;
}

/* According to the official specification, all strings in the torrent
 * file are supposed to be UTF-8 encoded. However, there are
 * non-compliant torrents around... If we encounter an invalid UTF-8
 * character, we assume it is ISO 8859-1 and convert it to UTF-8.
 */
static char* strcatUTF8( apr_pool_t* pool, char * s, char * append ) {
    char * p;
    char * rv = apr_palloc(pool, strlen(s) + (strlen(append) * 2));
    char * rvp = rv;
    memcpy(rv, s, strlen(s));

    /* Go to the end of the destination string */
    while(*rvp)
        rvp++;

    /* Now start appending, converting on the fly if necessary */
    for(p=append;*p;) {
        if(!(p[0] & 0x80)) {
            /* ASCII character */
            *(rvp++) = *(p++);
            continue;
        }

        if((p[0] & 0xE0) == 0xC0 && (p[1] & 0xC0) == 0x80) {
            /* 2-bytes UTF-8 character */
            *(rvp++) = *(p++); *(rvp++) = *(p++);
            continue;
        }

        if( ( p[0] & 0xF0 ) == 0xE0 && ( p[1] & 0xC0 ) == 0x80 &&
            ( p[2] & 0xC0 ) == 0x80 )
        {
            /* 3-bytes UTF-8 character */
            *(rvp++) = *(p++); *(rvp++) = *(p++);
            *(rvp++) = *(p++);
            continue;
        }

        if( ( p[0] & 0xF8 ) == 0xF0 && ( p[1] & 0xC0 ) == 0x80 &&
            ( p[2] & 0xC0 ) == 0x80 && ( p[3] & 0xC0 ) == 0x80 )
        {
            /* 4-bytes UTF-8 character */
            *(rv++) = *(p++); *(rv++) = *(p++);
            *(rv++) = *(p++); *(rv++) = *(p++);
            continue;
        }

        /* ISO 8859-1 -> UTF-8 conversion */
        *(rv++) = 0xC0 | ( ( *p & 0xFF ) >> 6 );
        *(rv++) = 0x80 | ( *(p++) & 0x3F );
    }
    
    return rv;
}
